#!/usr/bin/perl 
#**************************************************************************
#
# sherpa v0.1.4 - a system security configuration tool   
#
# Copyright (C) 2000 Rick Crelia (rick@nbank.net)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#**************************************************************************
# Local configuration options...modify as needed
#**************************************************************************

$sherpa_home="/usr/local/sherpa";  	# Where sherpa lives

$config_loc="/etc";			# Location of common system config
					# files

$log_dir="/var/log"; 			# Location of system log files

$perms_list="$sherpa_home/perms.lst";	# Location of file permission list

$sherpa_webroot="/usr/local/apache/htdocs/sherpa"; 
#$sherpa_webroot="/home/httpd/html/sherpa"; 	# This is the directory
					   	# where the HTML version
					   	# of sherpa's output will
					   	# go. Used only when $html=1

$sherpa_log="$log_dir/sherpa";		# log file location

# What services should sherpa evaluate?

# These are potentially risky services, but often come with Unix/Linux
# in an active state. sherpa will check for their usage and report it
@badsvcs = ("echo", "discard", "daytime", "smtp", "exec", 
            "chargen", "ntalk", "dtalk", "uucp", "tftp", 
            "bootps", "finger", "systat", "netstat", "linuxconf"); 

# These are used in report generation only
$today=`date`;
$ftoday=`date +%y%m%d`; chop($ftoday);
$host=$ENV{"HOSTNAME"};

$dirs2find = "$sherpa_home/find.lst";     # list of directories to search
					  # for finding world-writable,
					  # SGID, SUID and .rhosts files

#**************************************************************************
# End of local configuration options - do not modify below this point
# unless you know what you are doing...
#**************************************************************************

$ver="0.1.4";

$hr="-----------------------------------------------------------------\n\n";

use File::Find;
use Getopt::Std;


# ---------------------------------------------------------------

Main:

&parseopts();

&setup();

&check_sugids() if ($chksuids == 1);

&check_configs() if ($chkconfigs == 1);

&check_privs() if ($chkprivs == 1); 

&tidy();

close; 

exit;

# Subroutines follow

# -------------------------------------------------------------

sub tidy 
{
	  $finish = localtime(time);
	  $stop=time;
          $rtime = $stop - $begin;
	  $rmin = int($rtime/60);
	  $rsec = $rtime%60;		# modulus calculation
	  $msg = "$hr"."[$finish] - sherpa finished\n\n";
	  &prtmsg();
	  print LOG $msg if ($logging == 1);
          $msg = "\nRunning time = $rmin minutes $rsec seconds\n";
	  &prtmsg();
	  $msg = "\n-=< Results are in $sherpa_outfile >=-\n\n";
	  &prtmsg();
          if ($html == 1) 
          {
            &html_finish();
          }
	  close; exit; 
}

# -------------------------------------------------------------

sub check_privs 
{

    $now = localtime(time);
    $msg = "$hr"."[$now] - sherpa checking file and directory permissions\n";
    &prtmsg();
    &perm_chk();

    $now = localtime(time);
    $msg = "$hr"."[$now] - sherpa checking for world writable files\n";
    &prtmsg();
    &wwrite();

    $msg="\n";
    &prtmsg();
	

}

#----------------------------------------------------------------------

sub wwrite
{
    $wwcntr = 0; 
    find(\&findww, @fndlist);
    if ($wwcntr > 0) 
    {
      $msg = "\n$wwcntr world-writable files found:\n\n";
      &prtmsg();

      for ($i = 0; $i < $wwcntr; $i++)
      {
       $j = $i + 1;
       $j = " ".$j if ($j < 10); 
       $msg="$j\. $wwlist[$i]\n";
       &prtmsg();
      }
     }
     else
     {
      $msg = "\nNo world-writable files found. Good!\n\n";
      &prtmsg();
     }	
}

#----------------------------------------------------------------------

sub findww {
   if (
       (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) &&
       (-f $_) &&
       (! -l $_) &&
       (($mode & 02) == 02)
      )
   {
    push(@wwlist,$File::Find::name); $wwcntr++;
   }

}

#----------------------------------------------------------------------

sub perm_chk
{
   open PERMS,"$perms_list" || die "Cannot open permissions list!\n";
PERMCHK: 
   while ($line = <PERMS>) 
   {
     if ($line =~ /^\#/ || $line =~ /^$/) 
     {
       next PERMCHK;
     }
     else 
     {
        chop($line);
        ($target,$perms,$uname,$gname) = split(/,/,$line);
	$uid = getpwnam $uname; $gid = getgrnam $gname;
        if (!-e $target) 
        {
           $msg="$target not found!\n"; 
    	   &prtmsg();
	   next PERMCHK;
       }

	&get_fileinfo($target); 

        if ($found_perms != $perms) 
        {
           $msg="$target has wrong permissions =  $found_perms vs. $perms\n";
    	   &prtmsg();
           if ($fix==1) 
           {
             $msg="Changing permissions for $target to $perms\n";
    	     &prtmsg();
	     chmod oct($perms), $target || do
		{ 
		  $msg="Can't change permissions for $target to $perms\n";
    	   	  &prtmsg();
	        }
           }
        }
	else
	{
	  $msg="$target has correct permissions = $found_perms\n";	  
	  &prtmsg() if ($debug == 1);
	}
	
	if ($found_uid != $uid || $found_gid != $gid)
        {
          $msg="$target has wrong ownership - UID = $found_uname " . 	                       "GID = $found_gname\n";
	   &prtmsg();
           if ($fix==1) 
           {
             $msg="Changing ownership for $target to UID = $uname and " .
                   "GID = $gname\n";
	     &prtmsg();
	     chown $uid, $gid, $target  || do
		{ $msg="Can't change owner for $target to UID " .
                       "$uname GID $gname\n";
	   	  &prtmsg();
	        }
           }
        }
	else
	{
	  $msg="$target has correct ownership - UID = $uname " .
               "and GID = $gname\n";	  
	  &prtmsg() if ($debug == 1);
	}
	

    }

  }
  close(PERMS);
  $msg="\n"; &prtmsg();
 }

# -------------------------------------------------------------

sub check_configs 
{

        $now = localtime(time);
        $msg = "$hr"."[$now] - sherpa checking config files in $config_loc\n";
	&prtmsg();

	## Network security checks
        $now = localtime(time);
        $msg = "[$now] - \tchecking network security\n";
	&prtmsg();

	&check_hosts_equiv();

	&check_rhosts();

	&check_inetd();
        
	&check_pass();

        $msg="\n"; &prtmsg();
        
}

#----------------------------------------------------------------------
 
  sub check_inetd
  {
    $msg="\n"; &prtmsg();
    $target = "$config_loc/inetd.conf";
    open(INETD, $target) || die "Cannot open $target\n";
SVCCHK:
    while ($line=<INETD>) 
    {
     if ($line =~ /^\#/ || $line =~ /^$/) 
     {
	next SVCCHK;    #skip..service has been commented out	     
     }
     else 
     {
	chop($line);
	($svc,$stype,$sprot,$option,$user,$d1,$d2) = split /\s+/, $line, 7;
	if ($svc ne $lastsvc)
        {
          push(@services,$svc);  # avoid duplicates 
	  $svccntr++;
          if ($d1 eq "/usr/sbin/tcpd")
          {
            push(@wrapped,$svc);
	    $wrapcntr++;
          }
        }
        $lastsvc=$svc;
     }    
    } 
    close(INETD);
    $msg="Number of active services: $svccntr\n";
    &prtmsg();
    $msg="Number of wrappered active services: $wrapcntr\n";
    &prtmsg();
    $msg="Active services found (in $config_loc/inetd.conf):\n\n";
    &prtmsg();
    for ($i = 0; $i < $svccntr; $i++)
     {
       $j = $i + 1;
       $j = " ".$j if ($j < 10); 
       $msg="\t$j\. $services[$i]\n";
       &prtmsg();

       #check for problem svcs
       for ($k = 0; $i < $svcs2chk; $k++)
        {
          if ($services[$i] =~ /\$badsvc[\$k]/)
           {
	     push(@riskysvc,$badsvc[$k]);
             $badsvccntr++; 
           }
        }  
      }

    if ($badsvccntr > 0) 
    {
      $msg="\n";
      &prtmsg(); 
      for ($k = 0; $k < $badsvccntr; $k++) 
      {
         $msg = "Warning - $riskysvc[$k] can be risky to use\n";
         &prtmsg(); 
      }
    }

    $msg="\n";
    &prtmsg(); 
    $msg="Wrappered services found (in $config_loc/inetd.conf):\n\n";
    &prtmsg();
    for ($i = 0; $i < $wrapcntr; $i++)
     {
       $j = $i + 1;
       $j = " ".$j if ($j < 10); 
       $msg="\t$j\. $wrapped[$i]\n";
       &prtmsg();
     }

    if ($svccntr != $wrapcntr)
    {
         $msg="\n";
         &prtmsg(); 
         $msg = "Warning - Not all active services are wrappered!\n";
         &prtmsg(); 
         $msg = "If this includes internal services (e.g., echo, daytime)\n";
         &prtmsg(); 
         $msg = "then it's not a huge problem. However, consider turning\n";
         &prtmsg(); 
         $msg = "internal services off unless you really need them.\n";
         &prtmsg(); 
    }
    $msg="\n"; &prtmsg();
  } 


#---------------------------------------------------------------------

sub check_pass
{
        $now = localtime(time);
        $msg = "[$now] - sherpa checking password files for shadowing\n\n";
        &prtmsg();
        open PASSWD, "/etc/passwd" || die "Cannot open /etc/passwd!\n";
        $entry=<PASSWD>;
        if ($entry =~ /.*:x:/)
        {
                $msg = "\t User passwords are shadowed ... good!\n";
                &prtmsg();
        }
        else
        {
                $msg = "\t User passwords are NOT shadowed!\n";
                &prtmsg();
                if ($fix==1)
                {
                        $msg =  "\t Shadowing user passwords with pwconv...\n";
                        &prtmsg();
                        system "/usr/sbin/pwconv";
                }
                else
                {
                        $msg =  "Consider running pwconv to shadow your passwd file!\n";
                        &prtmsg();
                }
        }

        open GROUP, "/etc/group" || die "Cannot open /etc/group!\n";
        $entry=<GROUP>;
        if ($entry =~ /.*:x:/)
        {
                $msg = "\t Group passwords are shadowed ... good!\n";
                &prtmsg();
        }
        else
        {
                $msg = "\t Group passwords are NOT shadowed!\n";
                &prtmsg();
                if ($fix==1)
                {
                        $msg = "Shadowing group passwords with grpconv...\n";
                        &prtmsg();
                        system "/usr/sbin/grpconv";
                }
                else
                {
                        $msg =  "Consider running grpconv to shadow your group file!\n";
                        &prtmsg();
                }
        }
}                       



#---------------------------------------------------------------------

  sub check_rhosts   
  {
	find(\&find_rhosts, @fndlist);
        if ($rhostcntr > 0) 
	{
	  @sorted_rhosts = sort @rhosts; @rhosts = @sorted_rhosts;
          $msg = "\n** WARNING: Found $rhostcntr .rhosts files\n\n";
	  &prtmsg();
          for ($i = 0; $i < $rhostcntr; $i++)
          {
		&get_fileinfo($rhosts[$i]);
                $j = $i +1;
                $msg="$j\. $rhosts[$i] - Perms = $found_perms\n";
		&prtmsg();
          }
	}	
	else 
	{
	  $now = localtime(time);
	  $msg = "[$now] - \t   No .rhosts files found...good!\n";
	  &prtmsg();
	}
  	$msg="\n"; &prtmsg();
  } 

#---------------------------------------------------------------------

   sub find_rhosts 
   {
       if (/^\.rhosts$/ && -e) 
       { 
         push(@rhosts,$File::Find::name); 
         $rhostcntr++;
       }
   }  

#---------------------------------------------------------------------

  sub check_hosts_equiv 
  {
	# check for hosts.equiv
	$target = "$config_loc/hosts.equiv";
        if (-e $target) 
	{
	  $msg = "\n** WARNING: $target found!\n** Do you really need to " .
                 "establish trust with other systems?!?\n** See section " .
                 "2.3 of the AUSCERT Unix Security Checklist\n\n";
	  &prtmsg(); 
	  # check contents
	  open(HE,"$target");
	  while (<HE>) 
	  {
	     if (/^\+$/) 
	     {
		$msg = "** Found a + by itself on a line in $target... " .
		       "not a good thing!\n\n";
		&prtmsg();
	     }
	     if (/^\-/) 
	     {
		$msg = "** Found a - as first character in $target... " .
		       "not a good thing!\n** See CERT advisory [CA-91:12]\n\n";
		&prtmsg();
	     }
	     if (/\!/ || /\#/) 
	     {
		$msg = "** Found a comment character in $target... " .
		       "not a good thing!\n** Comments are not used in " .
		       "$target!\n\n";
		&prtmsg();
	     }

	  }		
		
  	  &get_fileinfo($target); $safe_perms = 600;
	  if ($found_perms!=$safe_perms) 
 	  {
		$msg = "$target has incorrect permissions - $found_perms \n";
		&prtmsg();	
		if ($fix==1) 
		{
	       	   chmod($safe_perms,$target);
		   $msg = "Changed permissions on $target to $safe_perms!\n";
		   &prtmsg();
        	}
       	  }
	  else 
	  {
		$msg = "Permissions on $target are $safe_perms ... good!\n\n";
		&prtmsg();
	  }
	} 
        else
        {
	  $now = localtime(time);
	  $msg = "[$now] - \t   No hosts.equiv file found...good!\n";
	  &prtmsg();
	}
    	 
  	$msg="\n"; &prtmsg();
  }


# -------------------------------------------------------------

sub check_sugids 
{

	$ugcntr = 0; $gcntr = 0; $ucntr = 0;
	$now = localtime(time);
	$msg = "$hr"."[$now] - sherpa checking for SUID and SGID files & dirs\n";
	&prtmsg();
	find(\&finduids, @fndlist);

	@sorted_uglist = sort @uglist; @uglist = @sorted_uglist;   
	$msg="\n$ugcntr SUID and SGID files found:\n\n";
	&prtmsg();
	for ($i = 0; $i < $ugcntr; $i++) 
	{
		$j = $i +1;
		$msg="$j\. $uglist[$i] is SUID and SGID\n";
		&prtmsg();
  	}

  	@sorted_ulist = sort @ulist; @ulist = @sorted_ulist;   
  	$msg="\n$ucntr SUID files found:\n\n";
	&prtmsg();
  	for ($i = 0; $i < $ucntr; $i++) 
  	{
		$j = $i +1;
		$msg="$j\. $ulist[$i] is SUID\n";
		&prtmsg();
  	}

  	@sorted_glist = sort @glist; @glist = @sorted_glist;
  	$msg="\n$gcntr SGID files found:\n\n";
	&prtmsg();	
  	for ($i = 0; $i < $gcntr; $i++) 
  	{
		$j = $i +1;
		$msg="$j\. $glist[$i] is SGID\n";
		&prtmsg();
  	}
  	$msg="\n"; &prtmsg();
}


# -------------------------------------------------------------

  sub finduids
  {
  
  if (-u && -g) 
   {
    push(@uglist,$File::Find::name); $ugcntr++;
   }
   elsif (-u && ! -g) 
   {
    push(@ulist,$File::Find::name); $ucntr++;
   }
   elsif (-g && ! -u) 
   {
    push(@glist,$File::Find::name); $gcntr++;
   }
   else 
   { #nada 
   }
 
  }


# -------------------------------------------------------------

sub setup  
{
	open(FLIST, $dirs2find);
	while ($dname = <FLIST>)
	{
	 chop($dname);
	 push(@fndlist,$dname);
	}
	close(FLIST);


        if ($html == 1) 
        {
	  $sherpa_outfile = "$sherpa_webroot/$ftoday.html";
	  open(OUT,"> $sherpa_outfile") || die "Cannot create $sherpa_outfile!";
          &html_begin();
        }
	else
	{
	  $sherpa_outfile = "$sherpa_home/sherpa.out";
	  open(OUT,"> $sherpa_outfile") || die "Cannot create $sherpa_outfile!";
          $msg = "sherpa $ver results - $host - $today\n\n";
   	  &prtmsg();	
        }

	$start = localtime(time);
	$begin=time;
	$msg = "[$start] - sherpa started\n";
   	&prtmsg();	
	if ($logging == 1) 
        {
	  if (-e $sherpa_log) 
	  {
	     open(LOG,">> $sherpa_log") || die "Cannot write to logfile!\n";
	  }
	  else 
	  {
	     open(LOG,"> $sherpa_log") || die "Cannot write to logfile!\n"; 
	  }
	  print LOG $msg;
        }	  
  	
	$msg="\n"; &prtmsg();

}


# -------------------------------------------------------------

sub get_fileinfo
{
     	$found_perms=sprintf("%o", (stat($_[0]))[2] & oct(7777));
	$found_uid=(stat($_[0]))[4];
	$found_gid=(stat($_[0]))[5];
	$found_perms; $found_uid; $found_gid; 
	$found_uname=getpwuid $found_uid;
        $found_gname=getgrgid $found_gid;
	$found_uname; $found_gname;
}


# -------------------------------------------------------------

sub prtmsg 
{
	print $msg if ($verbose==1);
	print OUT $msg;
}

# -------------------------------------------------------------

sub html_begin
{
  print OUT <<"HTML_START";
\n\n
<html>
<header>
<title>sherpa $ver results - $host - $today</title>
</header>
<body>
<h3>sherpa $ver results - $host - $today</h3>
<hr>
<p>
<pre>

HTML_START
}
	
# -------------------------------------------------------------


sub html_finish
{
  print OUT <<"HTML_END";

</pre>
<p>
<hr>
</body>
</html>
HTML_END
}

# -------------------------------------------------------------

sub parseopts
{

getopts("vhVdflwscpr:");

# 0 = minimal output to STDOUT, verbose to $sherpa_outfile 
# 1 = verbose all <default>
if ($opt_V) { $verbose=0; } else { $verbose=1; }

# extra output 0=no <default>, 1=yes
if ($opt_d) { $debug=1; } else { $debug=0; }

# Do you want the output to be an HTML page?
# 0=no <default>, 1=yes
if ($opt_w) { $html=1; } else { $html=0; }
		
# Should sherpa create a logfile ?
# 0=no  1=yes <default>
if ($opt_l) { $logging=0; } else { $logging=1; }

## Which checks do you want to run?
# defaults to yes in each instance

# check for SUID and SGID files/dirs
if ($opt_s) { $chksuids=1; } else { $chksuids=0; }

# check common config files
if ($opt_c) { $chkconfigs=1; } else { $chkconfigs=0; }

# check file & dir permissions and ownerships
if ($opt_p) { $chkprivs=1; } else { $chkprivs=0; }


# Do you want sherpa to correct problems that it can?
# 
# NOTE: The first time you run sherpa, leave this OFF and examine
# the output file to get an idea of what would be changed
# 0=no <default> 1=yes

if ($opt_f) 
{ 
  if (! $opt_p)
  {
   print "\n    You must specify -p to use -f\n\n";
   exit; 
  }
  $fix=1; 
} 
else 
{ 
  $fix=0; 
}

if ($opt_v) 
 {
   print "\nsherpa.pl v$ver - a system security configuration tool\n\n";
   exit;
 }

if ($opt_h)
 {
   print "\nsherpa.pl v$ver - a system security configuration tool\n\n";
   print "Usage: sherpa.pl [-hvVdwlscpf]

	-h  print this message
	-v  print version information
       	-V  turn off verbose output
	-d  print debugging output
	-w  generate output in HTML format
       	-l  turn logging off
       	-s  check for SUID/SGID files
       	-c  check system configuration files
	-p  check file/dir permissions in perms.lst
	-f  fix file and directory permissions (enforce perms.lst)	

     \n";
    exit; 
 }

if (!$opt_s && !$opt_c && !$opt_p)
 {
  print "\nYou did not specify any checks to perform.\n";
  print "Try -s, -c or -p (or some combination thereof)\n\n";
  print "For a list of options, type sherpa.pl -h\n\n";
  exit;
 } 
}
